! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

program fd4_demo_heat

  use fd4_mod
  implicit none
  include 'mpif.h'
  
  !! Setup parameters
  real, parameter :: radius = 0.5            ! rel. radius of initial heat pertubation
  integer, parameter :: grid(3) = 32         ! number of grid cells for x, y, z
  real(r_k), parameter :: ds(3) = 1.0        ! grid cell size for x, y, z
  real(r_k), parameter :: dt = 0.1           ! time step size
  integer, parameter :: nsteps = 1000        ! number of time steps to compute

  ! FD4 variable table
  type(fd4_vartab) :: vartab(1)
  integer, parameter :: THETA = 1
  ! FD4 domain
  type(fd4_domain), target :: domain
  integer :: dsize(3,2), bnum(3), nghosts(3)
  logical :: periodic(3)
  ! FD4 iterator
  type(fd4_iter) :: iter
  ! FD4 ghost communication
  type(fd4_ghostcomm) :: ghostcomm(2)
  ! FD4 netcdf communicator
#ifdef FD4_NETCDF4
  type(fd4_netcdf4_comm) :: nfcomm
#endif
  ! misc
  integer :: rank, nproc, err, bext(0:3), cabnd(3,2), caidxTheta, irank
  integer :: x, y, z, now, new, step
  real(r_k), allocatable :: buf(:,:,:,:)
  real(r_k) :: dtheta
  real :: global_pos(3), cr
  type(fd4_couple) :: couple
  real(r_k), pointer :: arraypointer(:,:,:)
  real(r_k), allocatable, target :: iniTheta(:,:,:)

  !! MPI Initialization
  call MPI_Init(err)
  call MPI_Comm_rank(MPI_COMM_WORLD, rank , err) 
  call MPI_Comm_size(MPI_COMM_WORLD, nproc, err) 

  !! Create the FD4 variable table
  !! (only one cell-centered variable 'theta' with 2 time steps)
  !                             name, nb, st,  unused, ini,      vthres,  discret.
  vartab(THETA) = fd4_vartab('theta',  1,  2, .false., 0.0, FD4_NOTHRES, FD4_CELLC )

  !! Create the FD4 domain
  dsize(1:3,1) = (/1, 1, 1/)    ! grid start indices
  dsize(1:3,2) = grid(1:3)      ! grid end   indices
  bnum(1:3) = grid(1:3) / 4     ! number of blocks in each dimension
  nghosts(1:3) = (/1, 1, 1/)    ! number of ghost cells in each dimension
  periodic(1:3) = .true.        ! periodic boundaries
  call fd4_domain_create(domain, bnum, dsize, vartab, nghosts, periodic, MPI_COMM_WORLD, err)
  if(err/=0) then
    write(*,*) rank, ': fd4_domain_create failed'
    call MPI_Abort(MPI_COMM_WORLD, 1, err)
  end if
  
  !! Allocate the blocks of the domain
  call fd4_util_allocate_all_blocks(domain, err)

  !! Allocate the buffer array for a single block with ghost cells
  call fd4_domain_max_bext(domain, bext(1:3), .true.)
  bext(0) = 1 ! 4th dimension not used here
  allocate( buf(bext(0),0:bext(1)-1,0:bext(2)-1,0:bext(3)-1) )
  buf = 0.0_r_k
  
  !! Initialize time step indicators
  now = 1
  new = 2

  !! Create couple context to put initial data to the FD4 domain
  call fd4_couple_create(couple, domain, err)
  
  ! add one couple array per rank for theta
  do irank = 0, nproc-1
    ! Set cabnd(:,:) to the bounds of the coupling arrays of current rank.
    ! (here we simply calculate bounds for an 1D-partitioning)
    cabnd(1:3,1) = (/   1+(irank*dsize(1,2))/nproc,          1,          1 /)
    cabnd(1:3,2) = (/ ((irank+1)*dsize(1,2))/nproc, dsize(2,2), dsize(3,2) /)
    ! Add a couple array of current rank with given bounds. Associate it with variable
    ! index THETA and the time step index now. FD4 assigns an index to local couple
    ! arrays (i.e. those that belong to the calling rank) to caidxTheta.
    call fd4_couple_add_array(couple, irank, THETA, now, cabnd, caidxTheta, err)
  end do 

  ! set this process' couple array bounds
  cabnd(1:3,1) = (/   1+(rank*dsize(1,2))/nproc,          1,          1 /)
  cabnd(1:3,2) = (/ ((rank+1)*dsize(1,2))/nproc, dsize(2,2), dsize(3,2) /)
  
  ! allocate the local couple array and add it to the couple context
  allocate(iniTheta(cabnd(1,1):cabnd(1,2),cabnd(2,1):cabnd(2,2),cabnd(3,1):cabnd(3,2)))
  iniTheta = 0.0_r_k
  arraypointer => iniTheta(:,:,:)
  call fd4_couple_set_local_3D_array(couple, caidxTheta, arraypointer, (/1,1,1/), err)

  ! commit the couple context, FD4 now checks the couple arrays and prepares MPI data types
  call fd4_couple_commit(couple, err)

  ! fill the couple array with initial data
  do z=cabnd(3,1),cabnd(3,2)
    do y=cabnd(2,1),cabnd(2,2)
      do x=cabnd(1,1),cabnd(1,2)
        ! get global coordinates of this grid cell and scale to [0,1]
        global_pos(1:3) = REAL( (/x,y,z/) - 1 ) / REAL(dsize(1:3,2) - 1)
        ! distance from domain center to current grid cell
        cr = sqrt( (global_pos(1)-0.5)**2+(global_pos(2)-0.5)**2+(global_pos(3)-0.5)**2 )
        if(cr < radius) then
          iniTheta(x,y,z) = 2.0_r_k * cos(3.14159*cr/(2*radius))
        end if
      end do 
    end do 
  end do 
  
  ! finally put data to FD4's data structures
  call fd4_couple_put(couple, err)

  ! delete couple context and deallocate couple array
  call fd4_couple_delete(couple, err)
  deallocate(iniTheta)

  !! Create ghost communicator for variable THETA (one for each time level)
  call fd4_ghostcomm_create(ghostcomm(1), domain, 1, (/THETA/), (/1/), err)
  call fd4_ghostcomm_create(ghostcomm(2), domain, 1, (/THETA/), (/2/), err)

  !! Initialize output and write initial data
#ifdef FD4_NETCDF4
  call fd4_netcdf4_open(nfcomm, domain, 'out.nc', 1, (/THETA/), (/now/), err)  
  call fd4_netcdf4_write(nfcomm,err)
#endif
#ifdef VIS5D
  call fd4_vis5d_open(domain, 'out.v5d', nsteps/100+1, 1, (/THETA/), (/now/), err)  
  call fd4_vis5d_write(err)
#endif
  
  !! Time stepping loop
  do step=1,nsteps

    ! exchange ghost cells for time level 'now'
    call fd4_ghostcomm_exch(ghostcomm(now), err)
    
    ! iterate over all local blocks
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))

      ! get theta with ghost cells from current block
      call fd4_iter_get_ghost(iter, THETA, now, bext, buf)

      ! loop over block's grid cells
      do z=1,iter%cur%ext(3)
        do y=1,iter%cur%ext(2)
          do x=1,iter%cur%ext(1)
            !                           ( d2T   d2T   d2T )
            !  theta_new =  theta_now + ( --- + --- + --- ) * dt
            !                           ( dx2   dy2   dz2 )
            ! calculate dtheta
            dtheta = ( buf(1,x-1,y,z) + buf(1,x+1,y,z) - 2*buf(1,x,y,z) ) / (ds(1) * ds(1)) &
                   + ( buf(1,x,y-1,z) + buf(1,x,y+1,z) - 2*buf(1,x,y,z) ) / (ds(2) * ds(2)) &
                   + ( buf(1,x,y,z-1) + buf(1,x,y,z+1) - 2*buf(1,x,y,z) ) / (ds(3) * ds(3))
            ! set updated theta value
            iter%cur%fields(THETA,new)%l(1,x,y,z) = buf(1,x,y,z) + dtheta * dt
          end do
        end do
      end do    

      call fd4_iter_next(iter)
    end do
    
    !! Write output
    if(mod(step,100)==0) then
      if(rank==0) write(*,'(A,I5)') 'step ',step
#ifdef FD4_NETCDF4
      call fd4_netcdf4_write(nfcomm, err, st_opt=(/new/))
#endif
#ifdef VIS5D
      call fd4_vis5d_write(err, st_opt=(/new/))
#endif
    end if

    ! swap time step indicators
    now = 3 - now
    new = 3 - new

  end do
  
  !! Close output files
#ifdef FD4_NETCDF4
  call fd4_netcdf4_close(nfcomm, err)
#endif
#ifdef VIS5D
  call fd4_vis5d_close(err)
#endif

  !! Delete the ghost communicator and the domain, finalize MPI
  call fd4_ghostcomm_delete(ghostcomm(1))
  call fd4_ghostcomm_delete(ghostcomm(2))
  call fd4_domain_delete(domain)
  call MPI_Finalize(err)

end program fd4_demo_heat
