! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.


module advection

  use fd4_mod
  implicit none
  include 'mpif.h'
  private

  real, parameter :: radius = 0.1            ! rel. radius of initial concentration bubble
  real(r_k), parameter :: erase = 1.d-5
  real(r_k), parameter :: eps = tiny(erase)
  integer, parameter :: varU = 1, varV = 2, varW = 3, varC = 4

  real(r_k), allocatable :: buf(:,:,:,:)
  real(r_k), allocatable, target :: iniC(:,:,:), iniU(:,:,:), iniV(:,:,:)
  integer :: bext(0:3)
  
  public :: varU, varV, varW, varC
  public :: advection_compute, advection_init

contains

  !! Allocate buffer array and initialize the variables in the domain
  subroutine advection_init(domain, now)
    type(fd4_domain), intent(inout), target :: domain
    integer, intent(in) :: now
    
    integer :: x, y, z, err, nproc, rank
    integer :: irank, cabnd(3,2), dsize(3,2), caidxC, caidxU, caidxV
    real :: global_pos(3), cr
    real(r_k), pointer :: arraypointer(:,:,:)
    type(fd4_couple) :: couple

    !! Allocate the buffer array for a single block with ghost cells
    call fd4_domain_max_bext(domain, bext(1:3), .true.)
    bext(0) = 1 ! 4th dimension not used here
    allocate( buf(bext(0),0:bext(1)-1,0:bext(2)-1,0:bext(3)-1) )
    buf = 0.0_r_k

    call MPI_Comm_rank(MPI_COMM_WORLD, rank, err) 
    call MPI_Comm_size(MPI_COMM_WORLD, nproc, err) 
    
    dsize = domain%dbnd
    
    !! Create couple context to put initial data to the FD4 domain
    call fd4_couple_create(couple, domain, err)
    
    !! Add couple arrays of all processes to the couple context
    do irank = 0, nproc-1
      ! Set cabnd(:,:) to the bounds of the coupling arrays of current rank.
      ! (here we simply calculate bounds for an 1D-partitioning)
      cabnd(1:3,1) = (/   1+(irank*dsize(1,2))/nproc,          1,          1 /)
      cabnd(1:3,2) = (/ ((irank+1)*dsize(1,2))/nproc, dsize(2,2), dsize(3,2) /)
      ! Add a couple array of current rank with given bounds. Associate it with variable
      ! index varC and the time step index now. FD4 assigns an index to local couple
      ! arrays (i.e. those that belong to the calling rank) to caidxC.
      call fd4_couple_add_array(couple, irank, varC, now, cabnd, caidxC, err)
      cabnd(1,2) = cabnd(1,2) + 1
      call fd4_couple_add_array(couple, irank, varU, now, cabnd, caidxU, err)
      cabnd(1,2) = cabnd(1,2) - 1
      cabnd(2,2) = cabnd(2,2) + 1
      call fd4_couple_add_array(couple, irank, varV, now, cabnd, caidxV, err)
    end do

    ! set this process' couple array bounds
    cabnd(1:3,1) = (/   1+(rank*dsize(1,2))/nproc,          1,          1 /)
    cabnd(1:3,2) = (/ ((rank+1)*dsize(1,2))/nproc, dsize(2,2), dsize(3,2) /)
    
    ! allocate the local couple arrays and add them to the couple context
    allocate(iniC(cabnd(1,1):cabnd(1,2),cabnd(2,1):cabnd(2,2),cabnd(3,1):cabnd(3,2)))
    iniC = 0.0_r_k
    arraypointer => iniC(:,:,:)
    call fd4_couple_set_local_3D_array(couple, caidxC, arraypointer, (/1,1,1/), err)

    allocate(iniU(cabnd(1,1):cabnd(1,2)+1,cabnd(2,1):cabnd(2,2),cabnd(3,1):cabnd(3,2)))
    iniU = 0.0_r_k
    arraypointer => iniU(:,:,:)
    call fd4_couple_set_local_3D_array(couple, caidxU, arraypointer, (/1,1,1/), err)

    allocate(iniV(cabnd(1,1):cabnd(1,2),cabnd(2,1):cabnd(2,2)+1,cabnd(3,1):cabnd(3,2)))
    iniV = 0.0_r_k
    arraypointer => iniV(:,:,:)
    call fd4_couple_set_local_3D_array(couple, caidxV, arraypointer, (/1,1,1/), err)

    ! commit the couple context, FD4 now checks the couple arrays and prepares MPI data types
    call fd4_couple_commit(couple, err)

        
    ! fill the couple arrays with initial data
    do z=cabnd(3,1),cabnd(3,2)
      do y=cabnd(2,1),cabnd(2,2)
        do x=cabnd(1,1),cabnd(1,2)
          ! get global coordinates of this grid cell and scale to [0,1]
          global_pos(1:3) = REAL( (/x,y,z/) - 1 ) / REAL(dsize(1:3,2) - 1)
          ! distance from domain center to current grid cell
          cr = sqrt( (global_pos(1)-0.25)**2+(global_pos(2)-0.5)**2+(global_pos(3)-0.25)**2 )
          if(cr < radius) then
            iniC(x,y,z) = 2.0_r_k * cos(3.14159*cr/(2*radius))
          end if
        end do
      end do
    end do    

    do y=cabnd(2,1),cabnd(2,2)
      global_pos(2) = REAL( y - 1 ) / REAL(dsize(2,2) - 1)
      iniU(:,y,:) = (global_pos(2) - 0.5) * 0.5
    end do

    do x=cabnd(1,1),cabnd(1,2)
      global_pos(1) = REAL( x - 1 ) / REAL(dsize(1,2) - 1)
      iniV(x,:,:) = (0.5 - global_pos(1)) * 0.5
    end do
    
    ! finally put data to FD4's data structures
    call fd4_couple_put(couple, err)
    
    ! delete couple context and deallocate couple array
    call fd4_couple_delete(couple, err)
    deallocate(iniC, iniU, iniV)

  end subroutine advection_init

  !! Compute one time step of a very simplisitc advection scheme on the current block of the iter
  subroutine advection_compute(iter, now, new, dt)
    type(fd4_iter), intent(inout) :: iter
    integer, intent(in) :: now, new
    real(r_k), intent(in) :: dt
    
    integer :: x, y, z
    real(r_k) :: deltac
    real(r_k) :: u0, u1, v0, v1, w0, w1
    integer   :: u0s,u1s,v0s,v1s,w0s,w1s

    ! get concentration varC with ghost cells from current block
    call fd4_iter_get_ghost(iter, varC, now, bext, buf)

    ! loop over block's grid cells
    do z=1,iter%cur%ext(3)
      do y=1,iter%cur%ext(2)
        do x=1,iter%cur%ext(1)
          ! check if there is mass in this cell and neighbor cells
          if( buf(1,x,y,z) < eps ) then
            if(   buf(1,x-1,y,z) + buf(1,x+1,y,z) &
                + buf(1,x,y-1,z) + buf(1,x,y+1,z) &
                + buf(1,x,y,z-1) + buf(1,x,y,z+1) < eps ) cycle
          end if
          ! get wind values at cell faces
          u0  = iter%cur%fields(varU,1)%l(1,x  ,y,z)
          u1  = iter%cur%fields(varU,1)%l(1,x+1,y,z)
          v0  = iter%cur%fields(varV,1)%l(1,x,y  ,z)
          v1  = iter%cur%fields(varV,1)%l(1,x,y+1,z)
          w0  = iter%cur%fields(varW,1)%l(1,x,y,z  )
          w1  = iter%cur%fields(varW,1)%l(1,x,y,z+1)
          ! determine source cell of the wind from its sign
          ! this cell limits the maximum flow between two cells
          u0s = INT(-0.5_r_k-SIGN(0.5_r_k,u0))
          u1s = INT( 0.5_r_k-SIGN(0.5_r_k,u1))
          v0s = INT(-0.5_r_k-SIGN(0.5_r_k,v0))
          v1s = INT( 0.5_r_k-SIGN(0.5_r_k,v1))
          w0s = INT(-0.5_r_k-SIGN(0.5_r_k,w0))
          w1s = INT( 0.5_r_k-SIGN(0.5_r_k,w1))
          ! calculate inflow using interpolation between current and neighbor cell value
          ! and consider the limit from the source cell
          deltac = MIN( .5*buf(1,x-1,y,z) + .5*buf(1,x,y,z), buf(1,x+u0s,y,z) ) * u0 &
                 - MIN( .5*buf(1,x+1,y,z) + .5*buf(1,x,y,z), buf(1,x+u1s,y,z) ) * u1 &
                 + MIN( .5*buf(1,x,y-1,z) + .5*buf(1,x,y,z), buf(1,x,y+v0s,z) ) * v0 &
                 - MIN( .5*buf(1,x,y+1,z) + .5*buf(1,x,y,z), buf(1,x,y+v1s,z) ) * v1 &
                 + MIN( .5*buf(1,x,y,z-1) + .5*buf(1,x,y,z), buf(1,x,y,z+w0s) ) * w0 &
                 - MIN( .5*buf(1,x,y,z+1) + .5*buf(1,x,y,z), buf(1,x,y,z+w1s) ) * w1
          ! add inflow to current value
          iter%cur%fields(varC,new)%l(1,x,y,z) = buf(1,x,y,z) + deltac * dt
          if(iter%cur%fields(varC,new)%l(1,x,y,z) < erase) iter%cur%fields(varC,new)%l(1,x,y,z) = 0.0_r_k
        end do
      end do
    end do

  end subroutine advection_compute

end module advection


program fd4_demo_advection

  use fd4_mod
  use advection
  implicit none
  include 'mpif.h'
  
  !! Setup parameters
  integer, parameter :: grid(3) = 32         ! number of grid cells for x, y, z
  real(r_k), parameter :: dt = 0.1           ! time step size
  integer, parameter :: nsteps = 2000        ! number of time steps to compute
  integer, parameter :: omod = 100           ! number of steps between output

  ! FD4 variable table
  type(fd4_vartab) :: vartab(4)
  ! FD4 domain
  type(fd4_domain), target :: domain
  integer :: dsize(3,2), bnum(3), nghosts(3)
  logical :: periodic(3)
  ! FD4 iterator
  type(fd4_iter) :: iter
  ! FD4 ghost communication
  type(fd4_ghostcomm) :: ghostcomm(2)
  ! FD4 netcdf communicator
#ifdef FD4_NETCDF4
  type(fd4_netcdf4_comm) :: nfcomm
#endif
  ! FD4 load balancing statistics
  type (fd4_balance_statistics) :: stats
  ! misc
  integer :: rank, err, now, new, step
  
  !! MPI Initialization
  call MPI_Init(err)
  call MPI_Comm_rank(MPI_COMM_WORLD,rank ,err) 

  !! Create the FD4 variable table
  !                             name, nb, st,  unused, ini,      vthres,  discret.
  vartab(varU) = fd4_vartab(     'u',  1,  1, .false., .0 , FD4_NOTHRES, FD4_FACEX )
  vartab(varV) = fd4_vartab(     'v',  1,  1, .false., .0 , FD4_NOTHRES, FD4_FACEY )
  vartab(varW) = fd4_vartab(     'w',  1,  1, .false., .05, FD4_NOTHRES, FD4_FACEZ )
  vartab(varC) = fd4_vartab(     'c',  1,  2, .false., .0 , FD4_NOTHRES, FD4_CELLC )

  !! Create the FD4 domain
  dsize(1:3,1) = (/1, 1, 1/)    ! grid start indices
  dsize(1:3,2) = grid(1:3)      ! grid end   indices
  bnum(1:3) = grid(1:3) / 8     ! number of blocks in each dimension
  nghosts(1:3) = (/1, 1, 1/)    ! number of ghost cells in each dimension
  periodic(1:3) = .true.        ! periodic boundaries
  call fd4_domain_create(domain, bnum, dsize, vartab, nghosts, periodic, MPI_COMM_WORLD, err)
  if(err/=0) then
    write(*,*) rank, ': fd4_domain_create failed'
    call MPI_Abort(MPI_COMM_WORLD, 1, err)
  end if
  
  !! Set some load balancing parameters
  call fd4_balance_params (domain, method=FD4_PART_SFC_PARALLEL, opt_lbtol=0.95)
  
  !! Allocate the blocks of the domain
  call fd4_util_allocate_all_blocks(domain, err)
  
  now = 1
  new = 2

  !! Allocate buffer array and initialize the variables in the domain
  call advection_init(domain, now)
  
  !! Create ghost communicator for variable varC (one for each time level of varC)
  call fd4_ghostcomm_create(ghostcomm(1), domain, 1, (/varC/), (/1/), err)
  call fd4_ghostcomm_create(ghostcomm(2), domain, 1, (/varC/), (/2/), err)

  !! Initialize output and write initial data
#ifdef FD4_NETCDF4
  call fd4_netcdf4_open(nfcomm, domain, 'out.nc', 4, (/varU,varV,varW,varC/), &
                        (/1,1,1,now/), err)
  call fd4_netcdf4_write(nfcomm,err)
#endif
#ifdef VIS5D
  call fd4_vis5d_open(domain, 'out.v5d', nsteps/omod+1, 4, (/varU,varV,varW,varC/), &
                        (/1,1,1,now/), err)
  call fd4_vis5d_write(err)
#endif
  
  !! Time stepping loop
  do step=1,nsteps

    ! exchange ghost cells for time level 'now'
    call fd4_ghostcomm_exch(ghostcomm(now), err)
    
    ! iterate over all local blocks
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))

      ! time the start of computations on current block
      call fd4_iter_start_clock(iter)

      ! compute simplistic advection scheme on current block
      call advection_compute(iter, now, new, dt)

      ! time the end of computations on current block, set block weight
      call fd4_iter_stop_clock(iter, 0.0)

      call fd4_iter_next(iter)
    end do

    ! rebalance the workload based on the measured block weights
    call fd4_balance_readjust(domain, err, opt_stats=stats)
    
    ! write output
    if(mod(step,omod)==0) then
      if(rank==0) write(*,'(A,I5,A,F6.3)') 'step ',step,'   bal:',stats%last_measured_balance
#ifdef FD4_NETCDF4
      call fd4_netcdf4_write(nfcomm, err, st_opt=(/1,1,1,new/))
#endif
#ifdef VIS5D
      call fd4_vis5d_write(err, st_opt=(/1,1,1,new/))
#endif
    end if

    ! swap time step indicators
    now = 3 - now
    new = 3 - new

  end do
  
  !! Close output files
#ifdef FD4_NETCDF4
  call fd4_netcdf4_close(nfcomm, err)
#endif
#ifdef VIS5D
  call fd4_vis5d_close(err)
#endif

  !! Delete the ghost communicator and the domain, finalize MPI
  call fd4_ghostcomm_delete(ghostcomm(1))
  call fd4_ghostcomm_delete(ghostcomm(2))
  call fd4_domain_delete(domain)
  call MPI_Finalize(err)

end program fd4_demo_advection

